/*
  Test driving the AB using the iPad
  Steering by using a lookup table of values to turn at various
  angles between 0 and 90 degrees

  The iPad will send Accell values scaled by 256 over using A6 message
  as 3, 16 bit integers for x,y,z Accell values
  Y Accell will be used for steering

  Throttle level will be sent as an int by iPad using an A3 message.
  Value will be 0 to 128

  A1 message received with value of 16 is ESTOP ALL STOP
  A1 message received with value of 32 is Forward Speed
  A1 message received with value of 64 is Reverse Speed
  
  Add a message to get a command byte indicating which abdrive command
  to execute, drive_speed or drive_ramp or drive_goto etc
  Then have three 16 bit integers for x,y,tp  

  Added message to get 16 bit integer(int16_t) form iPad along with
   a byte indicating which iPad control sent the value

  Fixed bug in RxMessage() in last step getting checksum now goes back and
  waits for a new message if checksum is bad .

  Each message will be of the following format
  
  MsgId  byte    Just something made up, maybe lower nibble is length
  Len    byte    Number of bytes remaining in message after this byte including chksum byte
  Data0  byte    Data bytes what they mean is message dependent
   .
   .
  Datan  byte
  Checksum       Checksum byte, the 2's complement of the sum of all bytes except MsgId
  

  Two simple messages for this test

  A single data byte to simulate switch closure or a key press

  MsgId = 0xA1
  Len   = 2
  Databyte
  CheckSum

  Send three accell values by iPad Tech basic
 
  MsgId = 0xA6
  Len = 7
  lob ax
  hob ax
  lob ay
  hob ay
  lob az
  hob az
  CheckSum

  The following messages for multiple byte variables have the
  LOB Low Order Byte and HOB High Order Byte arrainged to work 
  with Tech Basic side packing of the message and having the right
  byte order.

  Message for sending and receiving a 16 bit integer(int16_t) form iPad
  Used to test byte order and plus and minus values

  MsgId = 0xA3
  Len = 4
  WhoByte
  LOB
  HOB
  CheckSum

   WhoByte = 0 Ping sensor inches
             1 Ping sensor  cm
             2 Ping sensor counts
             3 Left Wheel ticks
             4 Right Wheel ticks
             5 Throttle position received from iPad

  Message for getting abdrive commands to execute a specific sequence of
  movements for the Activity Bot

  MsgId = 0xA7
  Len = 8
  CmdByte
  LOB r
  HOB r
  LOB l
  HOB l
  LOB tp  pause time
  HOB tp
  CheckSum

  CmdByte = 1 Execute drive_speed(L,R) sequence
          = 2 Execute drive_goto(L,R) sequence
          = 3 Execute drive_ramp(L,R) sequence  
*/

#include "simpletools.h"                      // Include simple tools
#include "fdserial.h"
#include "abdrive.h"
#include "ping.h"

// Comment or UnComment the next line to print debug info to Terminal
//#define DEBUG

// Comment or UnComment the next line to move the bot
#define MOVBOT

#define MAXCHARS 78

// Serial connection of XBEE Wifi to Activity Board or Board of Education
// Comment or Uncomment to select one

//#define BOE
#define ABOT

#ifdef BOE       // XBEEwifi connected pins
#define RX 10
#define TX 9
#define PINGPIN 7
#endif

#ifdef ABOT    // XBEEwifi connected pins
#define RX 9
#define TX 8
#define PINGPIN 7
#endif

/* Parsing the incoming messages requires a state machine driven by
   each incoming byte from the serial port
   With the following states
*/ 

#define WAIT4MSGID 1      //Get bytes until we see one we know
#define GETLENBYTE 2      //Get the number of bytes remaining in message
#define GETDATABYTES 3    //Get the data bytes. Len byte - 1 bytes
#define GETCHECKSUM 4     //Get checksum byte

#define MAXMSGSIZE 16     //Small messages for now

/* Define a unions to overlay the data bytes from an incoming message for
   multibyte variables 

*/

union A3Message {
 uint8_t bytes[3];
 struct __attribute__ ((__packed__)) data3_t {
  uint8_t who;
  int16_t value;
 }data;
} A3MsgD;
 

/* Create a union to overlay the $A7 message values with an array of
   bytes to use in receiving the message and parsing out the 16 bit values
   A union uses the same memory locations for all fields in the union.
   Allowing the same memory locations to be defined in different ways.
   Because of this we need a Struct to define three seperate elements that
   we want the array of bytes to overlay.
*/

// Need to use __attribute__ ((__packed__)) or the fields of the struct will be padded
// by compiler and CMD will parse right but the three 16bit values will be off and not
// match the array of bytes. Basicly a pad byte will be added after CMD to make everything
// on an even boundry and now it does not properly overlay the Message array of bytes.

union A7Message {
 uint8_t bytes[7];
 struct __attribute__ ((__packed__)) data_t {   
  uint8_t CMD; 
  int16_t R;
  int16_t L;
  uint16_t tp;
 }data;
}A7MsgD;

/* Define the A6 message data which will receive iPad Accel values x,y,z
*/

union A6Message {
 uint8_t bytes[6];
 struct __attribute__ ((__packed__)) data6_t {
  int16_t Xacel;
  int16_t Yacel;
  int16_t Zacel;
 }data;
}A6MsgD;



                        

fdserial *xbwifi;        // Serial device to talk to XBEEwifi

// Define CMD values for various abdrive code sequences to execute
#define ABSPEED 1
#define ABGOTO  2
#define ABRAMP  3

// Seperate tx and rx message buffers this version
// An array to hold an incoming message of max length
uint8_t MsgBuffer[MAXMSGSIZE];
uint8_t MsgTxBuffer[MAXMSGSIZE];

// Message parsing variables
uint8_t rxchar;
uint8_t MyChkSum;
uint8_t BValue;         // Byte Value from a message
int16_t IValue;         // Integer Value from a message
uint16_t WValue;        // Word Value from a message

// ABDrive varaibles
uint8_t CmdValue;       // ABDrive Command
int16_t RValue;         // ABDrive Right Value
int16_t LValue;         // ABDrive Left Value
uint16_t TValue;        // Pause Time Value

//uint8_t *pRxMsgPtr;
//uint8_t *pMsgPtr;
uint8_t bytecount;
uint8_t bytesneeded;
int     RxMsgState;


//Constant array of ticks to be added to drive_speed to turn at
//various angles
#define DX 32
#define SHFTDX 5  //2^5 = 32
#define XROWS 9
#define XMAXROW 8
#define YROWS 3

const int16_t acel2angle[XROWS][YROWS] =
{
//X,R,L or
//X,L,R depending on sign of incoming Accel value
{0,0,0},     //Row 0
{32,3,-3},   //Row 1
{64,7,-6},
{96,10,-9},
{128,13,-13},
{160,16,-16},
{192,19,-19},
{224,23,-22},
{256,26,-25}
};

//Define a struct to return interpolated values from the table
typedef struct  {
 int16_t col1;
 int16_t col2;
}agltks_t;

// Speed var updated by iPad sending an A1 message with throttle position
#define MAXSPEED 128

int16_t speed = 0;    //Throttle position speed always 0 to MAXSPEED
int16_t abspeed = 0;  //Activity bot speed +/- speed

//Ticks added to left and right wheel for steering
int16_t LeftSteer = 0; 
int16_t RightSteer = 0;

//Forward Reverse multiply speed by +- 1
int16_t direction = 1;



/* For testing of the receive message function. Just define it here
   for later version launch it into it's own cog to watch for messages
   and handshake with main line code when message is available
*/
void RxMessage() {
uint8_t chktest;
uint8_t idx;      //Stand alone version only
int ch;

// HERE IN STAND ALONE TEST VERSION
int msgflag;      // Set to 1 to build a new message
                  // State Machine sets it to 0 indicating have a message
                  // When in seperate cog handshake flag will be used

  //Get a char from serial port wait for it to arrive
  //In this test version it will hang the system
  RxMsgState = WAIT4MSGID;  // HERE IN STAND ALONE TEST VERSION
  msgflag = 1;              // HERE IN STAND ALONE TEST VERSION
  idx = 0;                  // HERE IN STAND ALONE TEST VERSION
 
#ifdef DEBUG
 print("\nRxMessage Looking\n");
#endif
 while(msgflag == 1) {

  rxchar = fdserial_rxChar(xbwifi);
  // Execute state machine on rxchar
  switch (RxMsgState) {
   case WAIT4MSGID:
        if ( (rxchar == 0xA1) || (rxchar == 0xA7) || (rxchar == 0xA3)
             || (rxchar == 0xA6) ) {
#ifdef DEBUG
          print("\nGot Message %x\n",rxchar);
#endif
          MsgBuffer[idx++] = rxchar;
          MyChkSum = 0;
          RxMsgState = GETLENBYTE;
        }
        break;
   case GETLENBYTE:
        // Already pointing at len byte
        MsgBuffer[idx++] = rxchar;
        bytesneeded = (uint8_t)rxchar -1; // Do not include checksum byte
        bytecount = 0;
        MyChkSum = (uint8_t)rxchar;       // First byte of sum
        RxMsgState = GETDATABYTES;
        break;
   case GETDATABYTES:
        //Already pointing at next byte
        MsgBuffer[idx++] = rxchar;
        MyChkSum += (uint8_t)rxchar;
        bytecount++;
        if(bytecount == bytesneeded) {
         RxMsgState = GETCHECKSUM;
        }        
        break;
   case GETCHECKSUM:
        //Now pointing at checksum byte
        MsgBuffer[idx] = rxchar;  // No need to inc to next byte this time
        /*Adding the received checksum byte to MyChkSum should be
          zero if we have a valid checksum because the checksum in message
          is the negitive of the sum
        */
#ifdef DEBUG
        //print("Doing checksum %d  %d\n",MyChkSum,(int8_t)rxchar);
#endif
        chktest = (uint8_t)rxchar + MyChkSum;
        chktest = chktest & 0xFF;  //Only 8 bits wanted
        if(chktest == 0) {
         msgflag = 0;  // Were done have valid message
        }
#ifdef DEBUG
        else {
         print("Bad Checksum\n");
         print("\nRxMessage Looking Again");
        }
#endif 
        RxMsgState = WAIT4MSGID; // Redundent in stand alone version Bug fix V2 
        fdserial_rxFlush(xbwifi);               // Clear the serial buffer
        idx = 0;                // Start at beginning of buffer if bad chksum
        break;
  }

 }// End while msgflag

}

// Send 16bit values back to iPad using A3 message
// Send them one at a time rather than have a single large message defined
// with everthing because you can never think of everything.

void SendMsgA3(uint8_t who, int16_t ival) {
uint8_t chksum;

 //Populate the transmit message buffer
 MsgTxBuffer[0] = 0xA3;
 MsgTxBuffer[1] = 4;
 chksum = 4;                           // Include data len in checksum
 A3MsgD.data.who = who;
 A3MsgD.data.value = ival;
 for(int i=2;i<5;i++){
  MsgTxBuffer[i] = A3MsgD.bytes[i-2];
  chksum = chksum + A3MsgD.bytes[i-2];
 }

 //Compute the final checksum as 2's complement of sum
 chksum = (~chksum +1) & 0xFF;
 MsgTxBuffer[5] = chksum;
#ifdef DEBUG
  print("Message A3\n");
  print("Who = %d Ival = %d\n",A3MsgD.data.who,A3MsgD.data.value);
  for(int m=0;m<6;m++) {
   print("$%2x | ",MsgTxBuffer[m]);
  }
  print("\n");
#endif

 //Send the message
 fdserial_txFlush(xbwifi);  //Clear the fdserial Tx buffer
 for(int x= 0;x<6;x++){
  fdserial_txChar(xbwifi,MsgTxBuffer[x]);
 } 

}


//Function to execute abdrive sequences based on command byte
void ExecuteABdrive(uint8_t BV,int16_t XV,int16_t YV,uint16_t TV) {

  switch(BV) { // Case on command byte
    case ABSPEED:
       drive_speed(LValue,RValue);
       pause(TValue);
       drive_speed(0,0);
       break;
    case ABGOTO:
       drive_goto(LValue,RValue);
       pause(TValue);
       break;
    case ABRAMP:
       drive_ramp(LValue,RValue);
       pause(TValue);
       drive_ramp(0,0);
       break;
  }// end case

}

/* Interpolate steering angles table 
   Uses table acel2angle[9][3] and returns values from
   column 1 and 2 to  be added to R,L or L,R of speed to turn
   column 0 of table is x axis
NOTE: xval is always positive 
*/
agltks_t lkupTicks(int16_t xval) {
agltks_t lkups;

  //First see if we are at or over 1g(256) for xval
  if(xval >= 256)  {
   //Just return top of table
   lkups.col1 = acel2angle[XMAXROW][1];
   lkups.col2 = acel2angle[XMAXROW][2];
  }
  else {
   //Get the first table row the xval is between
   //Divide xval by delta x in table to get first row int
   //division truncates fractional part
   int XR1 = xval / DX;
   int XR2 = XR1 + 1;   //Next row up, val is between rows XR2 and XR1

   //Because the delta is a power of two we can do a shift to divide by X2-X1
   // Interpolate to get column 1 ticks to add
   // (x - X1) * (Y2 - Y1)
   lkups.col1 = (xval - acel2angle[XR1][0]) * (acel2angle[XR2][1] - acel2angle[XR1][1]);
   // Divide by (X2 - X1) by shifting right by table delta pwr of 2 
   lkups.col1 = lkups.col1 >> SHFTDX;
   // y = Y1 + lkups.col1
   lkups.col1 = acel2angle[XR1][1] + lkups.col1;
   // Interpolate to get column 2 ticks to add
   // (x - X1) * (Y2 - Y1)
   lkups.col2 = (xval - acel2angle[XR1][0]) * (acel2angle[XR2][2] - acel2angle[XR1][2]);
   // Divide by (X2 - X1) by shifting right by table delta pwr of 2 
   lkups.col2 = lkups.col2 >> SHFTDX;
   // y = Y1 + lkups.col1
   lkups.col2 = acel2angle[XR1][2] + lkups.col2;        
  }
 return lkups;
}


void ExecuteSteering(int16_t xa,int16_t ya,int16_t za) {
agltks_t aglcnts;
int clockwise;
static int prevTblLoc = 0;

 // See which table location we are now in
 int TblLoc = ya / DX;
 //Only add steering ticks if speed > 0 and we changed table locations
 if((speed > 0) && (TblLoc != prevTblLoc)) {
  //Save new table location
  prevTblLoc = TblLoc;
  //See if we turned clockwise or counterclockwise
  //ya values are negitive for turning iPad to the right for clockwise turns
  if (ya < 0) {
   //Clockwise change ya to positive value
   clockwise = 1;
   int16_t tmpya = ~ya + 1;
   aglcnts = lkupTicks(tmpya);
  }
  else {
   //Counter Clockwise
   clockwise = 0;
   aglcnts = lkupTicks(ya);
  }
  // Use Table columns 1 and 2 R,L or L,R depending on which way turning
  if(clockwise) {
   LeftSteer = aglcnts.col1;
   RightSteer = aglcnts.col2;
  }
  else {
   LeftSteer = aglcnts.col2;
   RightSteer = aglcnts.col1;
  }
#ifdef DEBUG
   print("LeftSteer = %d RightSteer = %d\n",LeftSteer,RightSteer);
#endif
  //When in Reverse speed is negative so we need to use var direction to 
  //flip the sign of the steering adjustments so when wanting to add speed to
  //a wheel we do add a negitive to a negitive
  //direction is plus or minus 1 depending on forward or reverse
  drive_ramp(abspeed +  LeftSteer, abspeed + RightSteer);
 }
}



int main()   // Main function
{
uint8_t ix;  //Index into MsgBuffer array
int ch;
int leftW;
int rightW;
int pingin;

  // Add startup code here.
  xbwifi = fdserial_open(RX,TX,0,9600);   // Full Duplex serial to XBEE
  fdserial_rxFlush(xbwifi);               // Clear the serial buffer
  pause(100);

  RxMsgState = WAIT4MSGID;


 
  while(1)
  {
    // Add main loop code here.


    //Get a valid message, waits forever for one
    RxMessage();

    //Process the valid message
    //Have plenty of time not watching for new message 
    //until done.

    // Start at beginning of message
    ix = 0;
    
    //Use a case statement to process each type of message
    switch (MsgBuffer[ix++]) {
     case 0xA1:
       ix++;       // Bump index past len byte of message
       BValue = MsgBuffer[ix];
#ifdef DEBUG
       print("Message A1\n");
       print("Byte Value = %d\n",BValue);
#endif
       //Use the byte as if it was a keypress command
       switch (BValue) {
        case 1:
         high(26); // Turn on P26 Led
#ifdef DEBUG
         print("P26 ON\n");
#endif
         break;
        case 2:
         low(26);  // Turn off P26 Led
#ifdef DEBUG
         print("P26 OFF\n");
#endif
         break;
        case 4:
         high(27); // Turn on P27 Led
#ifdef DEBUG
         print("P27 ON\n");
#endif
         break;
        case 8:
         low(27); // Turn off P27 Led
#ifdef DEBUG
         print("P27 OFF\n");
#endif
         break;
        case 16: //All STOP
         drive_ramp(0,0);
         speed = 0;
         LeftSteer = 0;
         RightSteer = 0;
#ifdef DEBUG
         print("ESTOP\n");
#endif
         break;
        case 32:
         direction = 1; // Forward speed
#ifdef DEBUG
         print("FORWARD\n");
#endif
         break;
        case 64:
         direction = -1; // Reverse speed
#ifdef DEBUG
         print("REVERSE\n");
#endif
         break;
       }
       break;
     case 0xA6:
       ix++;       // Bump index past len byte of message
       //Copy the message bytes into the union
       for(int b=0;b<6;b++) {
        A6MsgD.bytes[b] = MsgBuffer[ix++];
       }
#ifdef DEBUG
       print("Message A6\n");
       print("X = %d Y = %d Z = %d\n",A6MsgD.data.Xacel,
                                      A6MsgD.data.Yacel,
                                      A6MsgD.data.Zacel);
#endif
#ifdef MOVBOT
       ExecuteSteering(A6MsgD.data.Xacel,A6MsgD.data.Yacel,A6MsgD.data.Zacel);
#endif
       break;
     case 0xA3:
       ix++;       // Bump index past len byte of message
       A3MsgD.bytes[0] = MsgBuffer[ix++];  //WHO
       A3MsgD.bytes[1] = MsgBuffer[ix++];  //LOB
       A3MsgD.bytes[2] = MsgBuffer[ix++];  //HOB
       BValue = A3MsgD.data.who;
       IValue = A3MsgD.data.value;
       if(A3MsgD.data.who == 5) { // Receiving throttle position, always positive number
         speed = A3MsgD.data.value;
         if(speed < 0) {
          speed = 0;
         }
         if(speed > MAXSPEED) {
          speed = MAXSPEED;
         }
         //Set abspeed depending on direction +/- 1
         abspeed = speed * direction;
#ifdef MOVBOT
         //Set new speed
         drive_ramp(abspeed + LeftSteer, abspeed + RightSteer);
#endif
       }
#ifdef DEBUG
       print("Message A3\n");
       print("Who = %d\n",BValue);
       print("ABSpeed = %d\n",abspeed);
#endif
       break;
     case 0xA7: // abdrive movement commands
       ix++;       // Bump index past len byte of message
       //Use the union to copy the  data bytes from the message
       A7MsgD.bytes[0] = MsgBuffer[ix++]; // Command byte
       A7MsgD.bytes[1] = MsgBuffer[ix++];
       A7MsgD.bytes[2] = MsgBuffer[ix++];
       A7MsgD.bytes[3] = MsgBuffer[ix++];
       A7MsgD.bytes[4] = MsgBuffer[ix++];
       A7MsgD.bytes[5] = MsgBuffer[ix++];
       A7MsgD.bytes[6] = MsgBuffer[ix++];
       CmdValue = A7MsgD.data.CMD;
       RValue = A7MsgD.data.R;
       LValue = A7MsgD.data.L;
       TValue = A7MsgD.data.tp;
#ifdef DEBUG
       print("Message A7\n");
       print("CMD = %d RValue = %d LValue = %d TValue = %d\n",
             CmdValue,RValue,LValue,TValue);
#endif
#ifdef MOVBOT
       ExecuteABdrive(CmdValue,RValue,LValue,TValue);
       drive_getTicks(&leftW,&rightW);
       // Get ping sensor distance
       pingin = ping_inches(PINGPIN);
       SendMsgA3(3,leftW);
       SendMsgA3(4,rightW);
//       SendMsgA3(4,12345);
       SendMsgA3(0,pingin);
#endif
       break;
   }// end switch message
    
  }  
}
